using UnityEngine;

public static class MeshFactory {

	private static float _meshWidth, _meshHeight;
	private static int _gridWidth, _gridHeight;

	// Creates a plane of the given size that is split into triangles arranged in a grid.
	// The size of one grid square is as close as possible to desiredGridSize,
	// while still keeping the grid uniform
	public static Mesh CreatePlane(float meshWidth, float meshHeight, float desiredGridSize)
	{
		int gridWidth = Mathf.RoundToInt(meshWidth / desiredGridSize);
		int gridHeight = Mathf.RoundToInt(meshHeight / desiredGridSize);
		return CreatePlane(meshWidth, meshHeight, gridWidth, gridHeight);
	}

	public static Mesh CreatePlane(float meshWidth, float meshHeight, int gridWidth, int gridHeight)
	{
		// These values need to be stored to use the GridCorner and GridUV methods
		_meshWidth = meshWidth;
		_meshHeight = meshHeight;
		_gridWidth = gridWidth;
		_gridHeight = gridHeight;

		// Create arrays for the mesh vertices and triangles
		int nVertices = (gridWidth + 1) * (gridHeight + 1);
		int nTriangles = gridWidth * gridHeight * 6;

		Vector3[] meshVertices = new Vector3[nVertices];
		Vector2[] meshUV = new Vector2[nVertices];
		int[] meshTriangles = new int[nTriangles];
		Vector3[] meshNormals = new Vector3[nVertices];
		Color[] meshColors = new Color[nVertices];

		// Generate the mesh vertices, uv and normals
		int index = 0;
		for (int y=0; y<=gridHeight; y++) {
			for (int x=0; x<=gridWidth; x++) {
				meshVertices[index] = GridCorner(x, y);
				meshUV[index] = GridUV(x, y);
				meshNormals[index] = Vector3.back;
				float value = Random.Range(0.9f, 1.0f);
				meshColors[index] = new Color(value, value, value);
				++index;
			}
		}

		// Generate the mesh triangles (every three entries are one triangle)
		index = 0;
		for (int y=0; y<gridHeight; y++) {
			for (int x=0; x<gridWidth; x++) {
				meshTriangles[index++] = GridVertexIndex(x  , y  );
				meshTriangles[index++] = GridVertexIndex(x  , y+1);
				meshTriangles[index++] = GridVertexIndex(x+1, y  );
				meshTriangles[index++] = GridVertexIndex(x+1, y  );
				meshTriangles[index++] = GridVertexIndex(x  , y+1);
				meshTriangles[index++] = GridVertexIndex(x+1, y+1);
			}
		}

		// Generate the mesh itself
		Mesh mesh = new Mesh();
		mesh.vertices = meshVertices;
		mesh.uv = meshUV;
		mesh.triangles = meshTriangles;
		mesh.normals = meshNormals;
		mesh.colors = meshColors;

		return mesh;
	}

	// Returns the position of a vertex in a grid.
	// The mesh width/height and grid width/height must be stored in the helper variables.
	private static Vector3 GridCorner(int x, int y) {
		return new Vector3(
			x * _meshWidth / _gridWidth - _meshWidth * 0.5f,
			y * _meshHeight / _gridHeight - _meshHeight * 0.5f,
			0.0f
		);
	}

	// Returns the UV-coordinates at the specified grid point.
	// This assumes that the texture is scaled to the size of the plane.
	// The grid width and height must be stored in the helper variables.
	private static Vector2 GridUV(int x, int y) {
		return new Vector2((float) x / _gridWidth, (float) y / _gridHeight);
	}

	// Returns the vertex index of the vertex at position (x,y) in the grid.
	// The grid width must be stored in the helper variables.
	private static int GridVertexIndex(int x, int y) {
		return y * (_gridWidth + 1) + x;
	}

}
