﻿using UnityEngine;
using System.Collections;

public class Player : MonoBehaviour {

	private bool grounded = false;
	private bool dead = false;

	private float jumpWallTouchTimer = 0.0f;
	private float jumpButtonTimer = 0.0f;
	private bool jumpButtonResetRequired = false;
	private Vector2 jumpDirection = Vector2.zero;
	private float horizontalTimeout = 0.0f;

	// Constants
	private static Vector2 wallJumpBoost = new Vector2(0.0f, 1.05f);
	private const float jumpStrengthHorizontal = 20.0f;
	private const float jumpStrengthVertical = 18f;

	private const float horizontalDeadzone = 0.3f;
	private const float horizontalDrag = 0.75f;
	private const float verticalDrag = 0.85f;

	private const float jumpActivationThreshold = 0.5f;
	private const float jumpReleaseThreshold = 0.3f;
	private const float jumpAnticipationDuration = 0.2f;
	private const float jumpRetainDuration = 0.1f;
	private const float jumpTimeoutDuration = 0.1f;

	private static int wallLayer;

	public GameObject bloodPrefab;
	public Vector2 respawnPoint;

	// Use this for initialization
	void Start()
	{
		wallLayer = LayerMask.GetMask("Walls");
	}
	
	// Update is called once per frame
	void FixedUpdate()
	{
		horizontalTimeout = Mathf.MoveTowards(horizontalTimeout, 0.0f, Time.deltaTime);
		jumpButtonTimer = Mathf.MoveTowards(jumpButtonTimer, 0.0f, Time.deltaTime);
		jumpWallTouchTimer = Mathf.MoveTowards(jumpWallTouchTimer, 0.0f, Time.deltaTime);

		float horizontal = Input.GetAxisRaw("Horizontal");
		float vertical = Input.GetAxisRaw("Vertical");

		// Horizontal movement
		if (horizontalTimeout < Mathf.Epsilon) {
			float desire = horizontal * 18.0f - rigidbody2D.velocity.x;
			if (desire * horizontal >= 0.0f) {
				float force = Mathf.Pow(Mathf.Abs(desire), 1.3f);
				rigidbody2D.AddForce(new Vector2(horizontal * force, 0.0f), ForceMode2D.Force);
			}
		}

		// Jump button press and release
		if (vertical > jumpActivationThreshold && !jumpButtonResetRequired) {
			jumpButtonResetRequired = true;
			jumpButtonTimer = jumpAnticipationDuration;
		} else if (vertical < jumpReleaseThreshold && jumpButtonResetRequired) {
			jumpButtonResetRequired = false;
		}

		// Horizontal drag
		if (Mathf.Abs(horizontal) < horizontalDeadzone && grounded) {
			rigidbody2D.velocity = new Vector2(rigidbody2D.velocity.x * horizontalDrag, rigidbody2D.velocity.y);
		}

		// Vertical jump stop
		if (vertical < jumpReleaseThreshold && rigidbody2D.velocity.y > 0.0f) {
			rigidbody2D.velocity = new Vector2(rigidbody2D.velocity.x, rigidbody2D.velocity.y * verticalDrag);
		}

		// Grounded and wall touch timer
		JumpChecks();

		// Jumping
		if (jumpWallTouchTimer >= Mathf.Epsilon && jumpButtonTimer >= Mathf.Epsilon) {
			jumpButtonTimer = 0.0f;
			jumpWallTouchTimer = 0.0f;
			rigidbody2D.velocity = new Vector2(rigidbody2D.velocity.x, 0.0f);
			Vector2 jumpVector = new Vector2(jumpStrengthHorizontal * jumpDirection.x, jumpStrengthVertical * jumpDirection.y);
			rigidbody2D.AddForce(jumpVector, ForceMode2D.Impulse);
			if (Mathf.Abs(jumpDirection.x) > 0.1f) horizontalTimeout = jumpTimeoutDuration;
		}

		Utilities.DebugRect(collider2D.bounds, jumpWallTouchTimer >= Mathf.Epsilon ? Color.green : Color.yellow);
	}

	void JumpChecks()
	{
		grounded = false;
		jumpDirection = Vector2.zero;

		// Ground
		RaycastHit2D[] hits = Physics2D.CircleCastAll(
			collider2D.bounds.center, collider2D.bounds.extents.x - 0.05f,
			new Vector2(0.0f, -1.0f), 0.1f, wallLayer
		);
		foreach (RaycastHit2D hitInfo in hits) {
			if (hitInfo.collider != collider2D) {
				grounded = true;
				jumpWallTouchTimer = jumpRetainDuration;
				jumpDirection += hitInfo.normal;
			}
		}

		if (grounded) {
			jumpDirection.Normalize();
			jumpDirection += wallJumpBoost;
			jumpDirection.Normalize();
			return;
		}

		// Right wall
		hits = Physics2D.CircleCastAll(
			collider2D.bounds.center, collider2D.bounds.extents.x - 0.05f,
			new Vector2(1.0f, 0.0f), 0.1f, wallLayer
		);
		foreach (RaycastHit2D hitInfo in hits) {
			if (hitInfo.collider != collider2D) {
				jumpWallTouchTimer = jumpRetainDuration;
				jumpDirection += hitInfo.normal;
			}
		}

		// Left wall
		hits = Physics2D.CircleCastAll(
			collider2D.bounds.center, collider2D.bounds.extents.x - 0.05f,
			new Vector2(-1.0f, 0.0f), 0.1f, wallLayer
			);
		foreach (RaycastHit2D hitInfo in hits) {
			if (hitInfo.collider != collider2D) {
				jumpWallTouchTimer = jumpRetainDuration;
				jumpDirection += hitInfo.normal;
			}
		}

		jumpDirection.Normalize();
		jumpDirection += wallJumpBoost;
		jumpDirection.Normalize();
	}

	void OnCollisionEnter2D(Collision2D collision)
	{
		if (collision.gameObject.CompareTag("Deadly")) Die();
	}

	void OnTriggerEnter2D(Collider2D other)
	{
		if (other.gameObject.CompareTag("Deadly")) Die();
	}

	public void Die()
	{
		if (dead) return;
		dead = true;
		collider2D.enabled = false;
		rigidbody2D.isKinematic = true;
		renderer.enabled = false;
		for (int i=0; i<12; i++) {
			GameObject obj = Instantiate(bloodPrefab, transform.position, Quaternion.identity) as GameObject;
			obj.transform.Rotate(0.0f, 0.0f, Random.Range(0.0f, 360.0f));
			obj.rigidbody2D.velocity = obj.transform.up * 3.0f;
		}
		StartCoroutine(Respawn());
	}

	IEnumerator Respawn()
	{
		transform.position = respawnPoint;
		yield return new WaitForSeconds(2.0f);
		SpriteRenderer render = (renderer as SpriteRenderer);
		render.enabled = true;
		Color c = render.color;
		c.a = 0.0f;
		while (c.a < 0.99f) {
			c.a += 0.05f * 1.8f;
			render.color = c;
			yield return new WaitForSeconds(0.05f);
		}
		c.a = 1.0f;
		render.color = c;
		rigidbody2D.isKinematic = false;
		collider2D.enabled = true;
		dead = false;
	}

}
